#include "crypt-port.h"

#if INCLUDE_sm3crypt

#include "alg-sm3.h"
#include "byteorder.h"

/*  Test vectors:

    "abc"
    SM3: 66c7f0f4 62eeedd9 d1f2d46b dc10e4e2 4167c487 5cf2f7a2 297da02b 8f4ba8e0

    "abcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcdabcd"
    SM3: debe9ff9 2275b8a1 38604889 c18e5a4d 6fdb70e5 387e5765 293dcba3 9c0c5732

    "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
    SM3: 639b6cc5 e64d9e37 a390b192 df4fa1ea 0720ab74 7ff692b9 f38c4e66 ad7b8c05

    "a" one million times
    SM3: c8aaf894 29554029 e231941a 2acc0ad6 1ff2a5ac d8fadd25 847a3a73 2b3b02c3

*/


#define ROTR(x, n)	((x << n) | (x >> (32 - n)))

#define R(i,A,B,C,D,E,F,G,H,T,Wi,Wj) do                               \
          {                                                           \
            SS1 = ROTR((ROTR((A), 12) + (E) + (T)), 7);         \
            SS2 = SS1 ^ ROTR((A), 12);                             \
            D += FF##i(A,B,C) + SS2 + ((Wi) ^ (Wj));                  \
            H += GG##i(E,F,G) + SS1 + (Wi);                           \
            B = ROTR((B), 9);                                      \
            F = ROTR((F), 19);                                     \
            H = P0 ((H));                                             \
          } while (0)

#define R1(A,B,C,D,E,F,G,H,T,Wi,Wj) R(1,A,B,C,D,E,F,G,H,T,Wi,Wj)
#define R2(A,B,C,D,E,F,G,H,T,Wi,Wj) R(2,A,B,C,D,E,F,G,H,T,Wi,Wj)

#define FF1(x, y, z)  (x ^ y ^ z)
#define FF2(x, y, z)  ((x & y) | (x & z) | (y & z))

#define GG1(x, y, z)  (x ^ y ^ z)
#define GG2(x, y, z)  ((x & y) | ( ~x & z))

/* Message expansion */
#define P0(x) ((x) ^ ROTR((x), 9) ^ ROTR((x), 17))
#define P1(x) ((x) ^ ROTR((x), 15) ^ ROTR((x), 23))

#define I(i)  ( w[i] = W[i] )
#define W1(i) ( w[i&0x0f] )
#define W2(i) ( w[i&0x0f] = P1(w[(i-16) & 0x0f] ^ w[(i-9) & 0x0f] ^ ROTR(w[(i-3) & 0x0f], 15)) \
                            ^ ROTR(w[(i-13) & 0x0f], 7) \
                            ^ w[(i-6) & 0x0f] )



static inline uint32_t buf_get_be32(const void *_buf)
{
  const unsigned char *in = _buf;
  return ((uint32_t)in[0] << 24) | ((uint32_t)in[1] << 16) | \
         ((uint32_t)in[2] << 8) | (uint32_t)in[3];
}

/*
 * Encode a length len*2 vector of (uint32_t) into a length len*8 vector of
 * (uint8_t) in big-endian form.
 */
static void
sm3_be32enc_vect(uint8_t * dst, const uint32_t * src, size_t len)
{
	/* Encode vector, two words at a time. */
	do {
		be32enc(&dst[0], src[0]);
		be32enc(&dst[4], src[1]);
		src += 2;
		dst += 8;
	} while (--len);
}

/*
 * Decode a big-endian length len*8 vector of (uint8_t) into a length
 * len*2 vector of (uint32_t).
 */
static void
sm3_be32dec_vect(uint32_t * dst, const uint8_t * src, size_t len)
{
	/* Decode vector, two words at a time. */
	do {
		dst[0] = be32dec(&src[0]);
		dst[1] = be32dec(&src[4]);
		src += 8;
		dst += 2;
	} while (--len);
}

static void
SM3_Transform(uint32_t state[static restrict 8],
    const uint8_t block[static restrict 64],
    uint32_t W[static restrict 64])
{
  uint32_t A, B, C, D, E, F, G, H, SS1, SS2;
  uint32_t w[16];

  static const uint32_t K[64] = {
    0x79cc4519, 0xf3988a32, 0xe7311465, 0xce6228cb,
    0x9cc45197, 0x3988a32f, 0x7311465e, 0xe6228cbc,
    0xcc451979, 0x988a32f3, 0x311465e7, 0x6228cbce,
    0xc451979c, 0x88a32f39, 0x11465e73, 0x228cbce6,
    0x9d8a7a87, 0x3b14f50f, 0x7629ea1e, 0xec53d43c,
    0xd8a7a879, 0xb14f50f3, 0x629ea1e7, 0xc53d43ce,
    0x8a7a879d, 0x14f50f3b, 0x29ea1e76, 0x53d43cec,
    0xa7a879d8, 0x4f50f3b1, 0x9ea1e762, 0x3d43cec5,
    0x7a879d8a, 0xf50f3b14, 0xea1e7629, 0xd43cec53,
    0xa879d8a7, 0x50f3b14f, 0xa1e7629e, 0x43cec53d,
    0x879d8a7a, 0x0f3b14f5, 0x1e7629ea, 0x3cec53d4,
    0x79d8a7a8, 0xf3b14f50, 0xe7629ea1, 0xcec53d43,
    0x9d8a7a87, 0x3b14f50f, 0x7629ea1e, 0xec53d43c,
    0xd8a7a879, 0xb14f50f3, 0x629ea1e7, 0xc53d43ce,
    0x8a7a879d, 0x14f50f3b, 0x29ea1e76, 0x53d43cec,
    0xa7a879d8, 0x4f50f3b1, 0x9ea1e762, 0x3d43cec5
  };

  A = state[0];
  B = state[1];
  C = state[2];
  D = state[3];
  E = state[4];
  F = state[5];
  G = state[6];
  H = state[7];

	sm3_be32dec_vect(W, block, 8);

  R1(A, B, C, D, E, F, G, H, K[0], I(0), I(4));
  R1(D, A, B, C, H, E, F, G, K[1], I(1), I(5));
  R1(C, D, A, B, G, H, E, F, K[2], I(2), I(6));
  R1(B, C, D, A, F, G, H, E, K[3], I(3), I(7));
  R1(A, B, C, D, E, F, G, H, K[4], W1(4), I(8));
  R1(D, A, B, C, H, E, F, G, K[5], W1(5), I(9));
  R1(C, D, A, B, G, H, E, F, K[6], W1(6), I(10));
  R1(B, C, D, A, F, G, H, E, K[7], W1(7), I(11));
  R1(A, B, C, D, E, F, G, H, K[8], W1(8), I(12));
  R1(D, A, B, C, H, E, F, G, K[9], W1(9), I(13));
  R1(C, D, A, B, G, H, E, F, K[10], W1(10), I(14));
  R1(B, C, D, A, F, G, H, E, K[11], W1(11), I(15));
  R1(A, B, C, D, E, F, G, H, K[12], W1(12), W2(16));
  R1(D, A, B, C, H, E, F, G, K[13], W1(13), W2(17));
  R1(C, D, A, B, G, H, E, F, K[14], W1(14), W2(18));
  R1(B, C, D, A, F, G, H, E, K[15], W1(15), W2(19));
  R2(A, B, C, D, E, F, G, H, K[16], W1(16), W2(20));
  R2(D, A, B, C, H, E, F, G, K[17], W1(17), W2(21));
  R2(C, D, A, B, G, H, E, F, K[18], W1(18), W2(22));
  R2(B, C, D, A, F, G, H, E, K[19], W1(19), W2(23));
  R2(A, B, C, D, E, F, G, H, K[20], W1(20), W2(24));
  R2(D, A, B, C, H, E, F, G, K[21], W1(21), W2(25));
  R2(C, D, A, B, G, H, E, F, K[22], W1(22), W2(26));
  R2(B, C, D, A, F, G, H, E, K[23], W1(23), W2(27));
  R2(A, B, C, D, E, F, G, H, K[24], W1(24), W2(28));
  R2(D, A, B, C, H, E, F, G, K[25], W1(25), W2(29));
  R2(C, D, A, B, G, H, E, F, K[26], W1(26), W2(30));
  R2(B, C, D, A, F, G, H, E, K[27], W1(27), W2(31));
  R2(A, B, C, D, E, F, G, H, K[28], W1(28), W2(32));
  R2(D, A, B, C, H, E, F, G, K[29], W1(29), W2(33));
  R2(C, D, A, B, G, H, E, F, K[30], W1(30), W2(34));
  R2(B, C, D, A, F, G, H, E, K[31], W1(31), W2(35));
  R2(A, B, C, D, E, F, G, H, K[32], W1(32), W2(36));
  R2(D, A, B, C, H, E, F, G, K[33], W1(33), W2(37));
  R2(C, D, A, B, G, H, E, F, K[34], W1(34), W2(38));
  R2(B, C, D, A, F, G, H, E, K[35], W1(35), W2(39));
  R2(A, B, C, D, E, F, G, H, K[36], W1(36), W2(40));
  R2(D, A, B, C, H, E, F, G, K[37], W1(37), W2(41));
  R2(C, D, A, B, G, H, E, F, K[38], W1(38), W2(42));
  R2(B, C, D, A, F, G, H, E, K[39], W1(39), W2(43));
  R2(A, B, C, D, E, F, G, H, K[40], W1(40), W2(44));
  R2(D, A, B, C, H, E, F, G, K[41], W1(41), W2(45));
  R2(C, D, A, B, G, H, E, F, K[42], W1(42), W2(46));
  R2(B, C, D, A, F, G, H, E, K[43], W1(43), W2(47));
  R2(A, B, C, D, E, F, G, H, K[44], W1(44), W2(48));
  R2(D, A, B, C, H, E, F, G, K[45], W1(45), W2(49));
  R2(C, D, A, B, G, H, E, F, K[46], W1(46), W2(50));
  R2(B, C, D, A, F, G, H, E, K[47], W1(47), W2(51));
  R2(A, B, C, D, E, F, G, H, K[48], W1(48), W2(52));
  R2(D, A, B, C, H, E, F, G, K[49], W1(49), W2(53));
  R2(C, D, A, B, G, H, E, F, K[50], W1(50), W2(54));
  R2(B, C, D, A, F, G, H, E, K[51], W1(51), W2(55));
  R2(A, B, C, D, E, F, G, H, K[52], W1(52), W2(56));
  R2(D, A, B, C, H, E, F, G, K[53], W1(53), W2(57));
  R2(C, D, A, B, G, H, E, F, K[54], W1(54), W2(58));
  R2(B, C, D, A, F, G, H, E, K[55], W1(55), W2(59));
  R2(A, B, C, D, E, F, G, H, K[56], W1(56), W2(60));
  R2(D, A, B, C, H, E, F, G, K[57], W1(57), W2(61));
  R2(C, D, A, B, G, H, E, F, K[58], W1(58), W2(62));
  R2(B, C, D, A, F, G, H, E, K[59], W1(59), W2(63));
  R2(A, B, C, D, E, F, G, H, K[60], W1(60), W2(64));
  R2(D, A, B, C, H, E, F, G, K[61], W1(61), W2(65));
  R2(C, D, A, B, G, H, E, F, K[62], W1(62), W2(66));
  R2(B, C, D, A, F, G, H, E, K[63], W1(63), W2(67));

  state[0] ^= A;
  state[1] ^= B;
  state[2] ^= C;
  state[3] ^= D;
  state[4] ^= E;
  state[5] ^= F;
  state[6] ^= G;
  state[7] ^= H;
}

/* Magic initialization constants. */
static const uint32_t initial_state[8] = {
	0x7380166f, 0x4914b2b9, 0x172442d7, 0xda8a0600,
	0xa96f30bc, 0x163138aa, 0xe38dee4d, 0xb0fb0e4e
};

/**
 * SM3_Init(ctx):
 * Initialize the SM3 context ${ctx}.
 */
void
SM3_Init(SM3_CTX * ctx)
{

	/* Zero bits processed so far. */
	ctx->count = 0;

	/* Initialize state. */
	memcpy(ctx->state, initial_state, sizeof(initial_state));
}


/**
 * SM3_Update(ctx, in, len):
 * Input ${len} bytes from ${in} into the SM3 context ${ctx}.
 */
static void
_SM3_Update(SM3_CTX * ctx, const void * in, size_t len,
    uint32_t tmp32[static restrict 72])
{
	uint32_t r;
	const uint8_t * src = in;

	/* Return immediately if we have nothing to do. */
	if (len == 0)
		return;

	/* Number of bytes left in the buffer from previous updates. */
	r = (ctx->count >> 3) & 0x3f;

	/* Update number of bits. */
	ctx->count += (uint64_t)(len) << 3;

	/* Handle the case where we don't need to perform any transforms. */
	if (len < 64 - r) {
		memcpy(&ctx->buf[r], src, len);
		return;
	}

	/* Finish the current block. */
	memcpy(&ctx->buf[r], src, 64 - r);
	SM3_Transform(ctx->state, ctx->buf, &tmp32[0]);
	src += 64 - r;
	len -= 64 - r;

	/* Perform complete blocks. */
	while (len >= 64) {
		SM3_Transform(ctx->state, src, &tmp32[0]);
		src += 64;
		len -= 64;
	}

	/* Copy left over data into buffer. */
	memcpy(ctx->buf, src, len);
}

/* Wrapper function for intermediate-values sanitization. */
void
SM3_Update(SM3_CTX * ctx, const void * in, size_t len)
{
	uint32_t tmp32[72];

	/* Call the real function. */
	_SM3_Update(ctx, in, len, tmp32);

	/* Clean the stack. */
	explicit_bzero(tmp32, 288);
}

static const uint8_t PAD[64] = {
	0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* Add padding and terminating bit-count. */
static void
SM3_Pad(SM3_CTX * ctx, uint32_t tmp32[static restrict 72])
{
	size_t r;

	/* Figure out how many bytes we have buffered. */
	r = (ctx->count >> 3) & 0x3f;

	/* Pad to 56 mod 64, transforming if we finish a block en route. */
	if (r < 56) {
		/* Pad to 56 mod 64. */
		memcpy(&ctx->buf[r], PAD, 56 - r);
	} else {
		/* Finish the current block and mix. */
		memcpy(&ctx->buf[r], PAD, 64 - r);
		SM3_Transform(ctx->state, ctx->buf, &tmp32[0]);

		/* The start of the final block is all zeroes. */
		memset(&ctx->buf[0], 0, 56);
	}

	/* Add the terminating bit-count. */
	be64enc(&ctx->buf[56], ctx->count);

	/* Mix in the final block. */
	SM3_Transform(ctx->state, ctx->buf, &tmp32[0]);
}
/**
 * SM3_Final(digest, ctx):
 * Output the SM3 hash of the data input to the context ${ctx} into the
 * buffer ${digest}.
 */
static void
_SM3_Final(uint8_t digest[32], SM3_CTX * ctx, uint32_t tmp32[static restrict 72])
{

	/* Add padding. */
	SM3_Pad(ctx, tmp32);

	/* Write the hash. */
	sm3_be32enc_vect(digest, ctx->state, 4);
}

/* Wrapper function for intermediate-values sanitization. */
void
SM3_Final(uint8_t digest[32], SM3_CTX * ctx)
{
	uint32_t tmp32[72];

	/* Call the real function. */
	_SM3_Final(digest, ctx, tmp32);

	/* Clear the context state. */
	explicit_bzero(ctx, sizeof(SM3_CTX));

	/* Clean the stack. */
	explicit_bzero(tmp32, 288);
}

/**
 * SM3_Buf(in, len, digest):
 * Compute the SM3 hash of ${len} bytes from ${in} and write it to ${digest}.
 */
void
SM3_Buf(const void * in, size_t len, uint8_t digest[32])
{
	SM3_CTX ctx;
	uint32_t tmp32[72];

	SM3_Init(&ctx);
	_SM3_Update(&ctx, in, len, tmp32);
	_SM3_Final(digest, &ctx, tmp32);

	/* Clean the stack. */
	explicit_bzero(&ctx, sizeof(SM3_CTX));
	explicit_bzero(tmp32, 288);
}
#endif /* INCLUDE_sm3crypt */